;;-*-Mode:emacs-lisp-*-
;; Copyright (c) 1990, Denys Duchier, Yale University.
;; This file is made available under the terms of the GNU General Public License.
;;
;; This code assumes that you compiled Epoch with DENYS_ALARM and DENYS_BUTTONS
;; defined. If ALL_MOTIONS was not defined, you won't get the advanced pointer
;; tracking feature (over dired, Buffer-menu, index, and rmail-summary buffers).
;;
;; This code is believed to be compatible with the default epoch lisp files. In
;; fact it requires that epoch.el be loaded.
;;
;; Put the following in your .emacs:
;;    (denys::install-nifty-select BUTTON1)
;;    (denys::install-aim-scroll   BUTTON2)
;; where you must substitute appropriate values for BUTTON1 and BUTTON2 (see
;; button definitions later in this file).

;;--the event handler. I think that explicit dispatching will be faster
;;--than looking up the appropriate handler in an alist. The reason I
;;--use variables to hold the handlers is that some handlers reset the
;;--values of other handlers to achieve context dependent effects.

(defvar denys::motion-handler          'denys::motion-normal)
(defvar denys::timer-handler           'denys::timer-normal)
(defvar denys::property-change-handler 'denys::property-change-normal)
(defvar denys::client-message-handler  'denys::client-message-normal)
(defvar denys::button-down-handler     'denys::button-down-normal)
(defvar denys::button-up-handler       'denys::button-up-normal)
(defvar denys::map-handler             'denys::map-normal)
(defvar denys::focus-handler           'denys::focus-normal)
(defvar denys::move-handler            'denys::move-normal)
(defvar denys::resize-handler          'denys::resize-normal)

(defun denys::event-handler ()
  (cond ((eq epoch::event-type 'motion)	         (funcall denys::motion-handler))
	((eq epoch::event-type 'timer)	         (funcall denys::timer-handler))
	((eq epoch::event-type 'property-change) (funcall denys::property-change-handler))
	((eq epoch::event-type 'client-message)	 (funcall denys::client-message-handler))
	((eq epoch::event-type 'button-down)	 (funcall denys::button-down-handler))
	((eq epoch::event-type 'button-up)	 (funcall denys::button-up-handler))
	((eq epoch::event-type 'map)	         (funcall denys::map-handler))
	((eq epoch::event-type 'focus)	         (funcall denys::focus-handler))
	((eq epoch::event-type 'move)	         (funcall denys::move-handler))
	((eq epoch::event-type 'resize)	         (funcall denys::resize-handler))
	(t (error "unknown event type: %s" epoch::event-type))))

;;--in case you don't use my style of event handling, here is a handler
;;--that will probably do the job.
(defun denys::prehistoric-event-handler ()
  (denys::pre-event-handler)
  (while (epoch::pop-event) (denys::event-handler))
  (denys::post-event-handler))

(setq epoch::event-handler
      (if (boundp 'epoch::pre-event-handler)
	  'denys::event-handler
	'denys::prehistoric-event-handler)
      epoch::event-handler-abort nil
      epoch::mouse-events t
      epoch::buttons-modify-buffer nil)

;;--also, pre and post processing maybe required. you may add new functions
;;--to the pre-event-handler-hook and post-event-handler-hook lists. For instance,
;;--it may be a good idea to set a global variable to indicate that you
;;--request some action to be performed, and have a post hook check that
;;--variable and perform the action only once (when exiting the event handling
;;--loop). eg. mouse-timeout in this file.

(defvar denys::pre-event-handler-hook nil)
(defvar denys::post-event-handler-hook nil)

;;--did we really go through the loop?
(defvar denys::went-through-event-loop-p t)

(defun denys::pre-event-handler ()
  (if (not denys::went-through-event-loop-p)
      (denys::mouse-abort))
  (run-hooks 'denys::pre-event-handler-hook)
  (setq denys::went-through-event-loop nil))

(setq epoch::pre-event-handler 'denys::pre-event-handler)

(defun denys::post-event-handler ()
  (run-hooks 'denys::post-event-handler-hook)
  (setq denys::went-through-event-loop t))

(setq epoch::post-event-handler 'denys::post-event-handler)

;;--when a task started with the mouse must be aborted, we need
;;--to restore the global state to normal. Each new funky task must
;;--add its own hooks to the following list (if any is required).
(defvar denys::mouse-abort-hook nil)

(defun denys::mouse-abort ()
  (run-hooks 'denys::mouse-abort-hook))

;;--hangs a function off a hook.
(defun denys::hang-off-hook (hook-name function)
  (if (memq function (symbol-value hook-name)) nil
    (set hook-name (cons function (symbol-value hook-name)))))

;;--this restores all event handlers to their default behaviour.
;;--we put the function on tyhe mouse-abort-hook list so that whenever
;;--we abort a mouse task, we restore all handlers to normal.

(defun denys::restore-handlers ()
  (setq denys::motion-handler          'denys::motion-normal
	denys::timer-handler           'denys::timer-normal
	denys::property-change-handler 'denys::property-change-normal
	denys::client-message-handler  'denys::client-message-normal
	denys::button-down-handler     'denys::button-down-normal
	denys::button-up-handler       'denys::button-up-normal
	denys::map-handler             'denys::map-normal
	denys::focus-handler           'denys::focus-normal
	denys::move-handler            'denys::move-normal
	denys::resize-handler          'denys::resize-normal))

(denys::hang-off-hook 'denys::mouse-abort-hook 'denys::restore-handlers)

;;--sometimes you want some event to abort the current task and resume
;;--the default behaviour. you can set handlers to the following function
;;--to achieve that effect.

(defun denys::mouse-restart ()
  (denys::mouse-abort)
  (denys::event-handler))

;;--------------------------------------------------------------------------------
;;  BUTTON HANDLERS
;;--------------------------------------------------------------------------------

;;--beep on undefined buttons
(defvar denys::mouse-ding-p t
  "*if non nil, ring the bell when button not handled.")

(defun denys::mouse-ding ()
  (if denys::mouse-ding-p (ding)))

;;--buttons are represented by numbers: left is 1, middle is 2, right is 3.
;;--to get a shifted button add 4, meta add 8, control add 16.
;;--whether the transition is up or down is specified by epoch::event-type
;;--which is either button-up or button-down.

(defconst denys::button-unknown 0)
(defconst denys::button-left    1)
(defconst denys::button-middle  2)
(defconst denys::button-right   3)

(defconst denys::button-shift-left   (+ 1 4))
(defconst denys::button-shift-middle (+ 2 4))
(defconst denys::button-shift-right  (+ 3 4))

(defconst denys::button-meta-left   (+ 1 8))
(defconst denys::button-meta-middle (+ 2 8))
(defconst denys::button-meta-right  (+ 3 8))

(defconst denys::button-control-left   (+ 1 16))
(defconst denys::button-control-middle (+ 2 16))
(defconst denys::button-control-right  (+ 3 16))

(defconst denys::button-meta-shift-left   (+ 1 4 8))
(defconst denys::button-meta-shift-middle (+ 2 4 8))
(defconst denys::button-meta-shift-right  (+ 3 4 8))

(defconst denys::button-control-shift-left   (+ 1 4 16))
(defconst denys::button-control-shift-middle (+ 2 4 16))
(defconst denys::button-control-shift-right  (+ 3 4 16))

(defconst denys::button-control-meta-left   (+ 1 8 16))
(defconst denys::button-control-meta-middle (+ 2 8 16))
(defconst denys::button-control-meta-right  (+ 3 8 16))

(defconst denys::button-control-meta-shift-left   (+ 1 4 8 16))
(defconst denys::button-control-meta-shift-middle (+ 2 4 8 16))
(defconst denys::button-control-meta-shift-right  (+ 3 4 8 16))

;;--maps for button transitions
(defvar denys::button-up-map   (make-vector 32 nil))
(defvar denys::button-down-map (make-vector 32 nil))

;;--temporaries
(defvar denys::tmp0 nil)
(defvar denys::tmp1 nil)
(defvar denys::tmp2 nil)

(defun denys::button-down-normal ()
  (if (setq denys::tmp0 (aref denys::button-down-map epoch::event-data))
      (funcall denys::tmp0)
    (denys::mouse-abort)
    (denys::mouse-ding)))

(defun denys::button-up-normal ()
  (if (setq denys::tmp0 (aref denys::button-up-map epoch::event-data))
      (funcall denys::tmp0)
    (denys::mouse-abort)
    (denys::mouse-ding)))

;;--------------------------------------------------------------------------------
;; UTILITIES
;;--------------------------------------------------------------------------------

;;--a handy function that does nothing and returns t.
(defun denys::noop () t)

;;--make sure rename-screen is available. default to noop.
(if (fboundp 'rename-screen) nil
  (fset 'rename-screen 'denys::noop))

;;--select epoch::event-window.
(defun denys::mouse-select ()
  (cond ((null epoch::event-window) nil)
	((eq epoch::event-window (selected-window)) t)
	(t (select-window epoch::event-window)
	   (rename-screen))))

;;--in window, move point to (x,y).
(defun denys::move-point-to-window-coordinates (x y window)
  (setq denys::tmp0 (window-hscroll window))
  (move-to-window-line y window)
  (move-to-column (if (> denys::tmp0 1) (+ x (1- denys::tmp0)) x) window)
  t)

;;--select window mouse is on and set point at mouse's location.
(defun denys::mouse-set-point ()
  (and (denys::mouse-select)
       (window-coordinates epoch::event-x epoch::event-y
			   epoch::event-window nil
			   'denys::tmp0 'denys::tmp1)
       (denys::move-point-to-window-coordinates denys::tmp0
						denys::tmp1
						epoch::event-window)))

;;--warp the mouse to the location of point. this is used when we
;;--cut text out of the buffer to keep the mouse over the new location
;;--of the point.
(defun denys::warp-mouse-to-point ()
  (setq denys::tmp0 (window-hscroll))
  (setq denys::tmp0 (if (> denys::tmp0 1)
			(- (current-column) (1- denys::tmp0))
		      (current-column)))
  (if (< denys::tmp0 0) (setq denys::tmp0 0))
  (setq denys::tmp1 (1- (count-lines (window-start) (point))))
  (if (bolp) (setq denys::tmp1 (1+ denys::tmp1)))
  (epoch::warp-mouse (+ denys::tmp0 (car denys::mouse-edges))
		     (+ denys::tmp1 (car (cdr denys::mouse-edges)))
		     nil))

;;--------------------------------------------------------------------------------
;; HIGHLIGHTING
;;--------------------------------------------------------------------------------

;;--highlighting is done using a special button. by default we assign attribute
;;--255 to it.

(defvar denys::mouse-button (make-button))
(defvar denys::mouse-attribute        255)
(defvar denys::mouse-edges            nil)
(defvar denys::mouse-before           nil)
(defvar denys::mouse-button-buffer    nil)
(defvar denys::mouse-button-start     nil)
(defvar denys::mouse-button-end       nil)
(defvar denys::mouse-button-requested nil)
(defvar denys::mouse-time             nil)
(defvar denys::mouse-paste-timeout    500)

(set-button-attribute denys::mouse-button denys::mouse-attribute)

;;--this removes the highlighting button from wherever it is.
(defun denys::remove-mouse-button ()
  (remove-button denys::mouse-button))

;;--remove highlighting button whenever we abort.
(denys::hang-off-hook 'denys::mouse-abort-hook 'denys::remove-mouse-button)

;;--one end of the highlighting button remaining anchored, move the
;;--other end to the point. which end remains anchored is determined
;;--by mouse-before, which tells us whether we where extending the
;;--highlight towards the beginning or the end of the buffer.

(defun denys::move-mouse-button-to-point ()
  (setq denys::mouse-button-requested t
	denys::mouse-button-buffer (current-buffer))
  (if denys::mouse-before
      (if (<= (point) denys::mouse-button-end)
	  (setq denys::mouse-button-start (point))
	(setq denys::mouse-button-start denys::mouse-button-end
	      denys::mouse-button-end (point)
	      denys::mouse-before nil))
    (if (>= (point) denys::mouse-button-start)
	(setq denys::mouse-button-end (point))
      (setq denys::mouse-button-end denys::mouse-button-start
	    denys::mouse-button-start (point)
	    denys::mouse-before t))))

;;--set point to mouse's location and extend highlight to match.
(defun denys::move-mouse-button ()
  (denys::mouse-set-point)
  (denys::move-mouse-button-to-point))

;;--in order to move the highlighting button, we set global variables
;;--and the post event handler will perform the actual move. thus if
;;--several moves are requested during event processing, only the last
;;--one need be performed. in other words, we get some kind of request
;;--compression for free.

(defun denys::mouse-button-hook ()
  (if denys::mouse-button-requested
      (progn
	(move-button denys::mouse-button
		     denys::mouse-button-start
		     denys::mouse-button-end
		     denys::mouse-button-buffer)
	(setq denys::mouse-button-requested nil))))

(denys::hang-off-hook 'denys::post-event-handler-hook 'denys::mouse-button-hook)

;;--------------------------------------------------------------------------------
;; NIFTY SELECT
;;--------------------------------------------------------------------------------

(defun denys::dired-double-click ()
  (denys::mouse-abort)
  (dired-find-file))

(defun denys::Buffer-menu-double-click ()
  (denys::mouse-abort)
  (Buffer-menu-this-window))

(defun denys::index-double-click ()
  (denys::mouse-abort)
  (index-other-window))

(defun denys::rmail-summary-double-click ()
  (denys::mouse-abort)
  (rmail-summary-goto-msg)
  (pop-to-buffer rmail-buffer))

(defvar denys::double-click-alist
  '((dired-mode         . denys::dired-double-click)
    (Buffer-menu-mode   . denys::Buffer-menu-double-click)
    (index-mode         . denys::index-double-click)
    (rmail-summary-mode . denys::rmail-summary-double-click))
  "*alist of elements of the form (major-mode . proc), where major-mode
is a symbol denoting a major mode, and proc is a function of no
arguments to be executed when you click twice (double click) in
the same location in a buffer with a corresponding major mode.
the default is to paste.")

(defvar denys::nifty-select-button nil)

;;--ignore the nifty button. ding on other buttons, but ignore them anyway.
(defun denys::button-nifty-ignore ()
  (if (not (= (% epoch::event-data 4) denys::nifty-select-button))
      (denys::mouse-ding)))

(defun denys::mouse-begin-nifty-select ()
  (denys::remove-mouse-button)
  (denys::mouse-set-point)
  (setq denys::nifty-select-button  (% epoch::event-data 4)       ;-remember button
	denys::button-down-handler 'denys::mouse-ding             ;-ignore other buttons
	denys::button-up-handler   'denys::button-up-maybe-select ;-wait for up event
	denys::motion-handler      'denys::motion-maybe-select    ;-in case we move before releasing button
	denys::focus-handler       'denys::focus-maybe-select     ;-in case we move out ...
	))

(defun denys::button-up-maybe-select ()
  (if (= (% epoch::event-data 4) denys::nifty-select-button)           ;-ignore other buttons
      (setq denys::button-down-handler 'denys::button-down-maybe-paste ;-watch button
	    denys::button-up-handler   'denys::mouse-abort             ;-orphan up event?
	    denys::motion-handler      'denys::mouse-restart           ;-in case we move
	    denys::mouse-time          (the-time))                     ;-keep track of the time
    (denys::mouse-ding)))

(defun denys::button-down-maybe-paste ()
  (cond ((or (not (= (% epoch::event-data 4) denys::nifty-select-button))    ;-another button was pressed
	     (> (- (the-time) denys::mouse-time) denys::mouse-paste-timeout)) ;-or waited too long
	 (denys::mouse-restart))
	((setq denys::tmp0 (assq major-mode denys::double-click-alist))
	 ;;--execute special double click function
	 ;;--we don't check that we are still in the same buffer because
	 ;;--the motion handler would have caught that.
	 (funcall (cdr denys::tmp0)))
	;;--default: 
	(t (setq denys::which-yank 'get-cut-buffer)
	   (denys::yank)
	   (setq denys::button-down-handler 'denys::yank
		 denys::button-up-handler   'denys::button-nifty-ignore))))

;;--one of: get-cut-buffer, yank, yank-pop
(defvar denys::which-yank 'get-cut-buffer)

;;--when yanking, first try the cut buffer, then cycle through the
;;--kill ring. skip head of kill ring if same as cut buffer.
(defun denys::yank ()
  (cond ((eq denys::which-yank 'get-cut-buffer)
	 (push-mark (point))
	 (setq denys::tmp0 (epoch::get-cut-buffer))
	 (if (and kill-ring (string-equal (car kill-ring) denys::tmp0))
	     (setq denys::which-yank 'yank-pop)
	   (setq denys::which-yank 'yank))
	 (insert denys::tmp0)
	 (setq denys::tmp0 nil)
	 (exchange-point-and-mark))
	((eq denys::which-yank 'yank)
	 ;;--delete insertion from cut buffer
	 (delete-region (point) (mark))
	 (rotate-yank-pointer 0)
	 (set-mark (point))
	 (insert (car kill-ring-yank-pointer))
	 (exchange-point-and-mark)
	 (setq denys::which-yank 'yank-pop))
	((eq denys::which-yank 'yank-pop)
	 ;;--delete previous insertion
	 (delete-region (point) (mark))
	 (rotate-yank-pointer 1)
	 (set-mark (point))
	 (insert (car kill-ring-yank-pointer))
	 (exchange-point-and-mark))
	(t (error "unexpected value of denys::which-yank: %s" denys::which-yank))))

;;--copy region. wait for other click requesting deletion.
;;--motion handler is set to mouse-restart.
(defun denys::button-down-maybe-copy ()
  (if (not (= (% epoch::event-data 4) denys::nifty-select-button))
      (denys::mouse-ding)
    (let ((last-command nil))
      (copy-region-as-kill (button-start denys::mouse-button)
			   (button-end   denys::mouse-button)))
    (epoch::store-cut-buffer (car kill-ring))
    (setq denys::button-down-handler 'denys::button-down-maybe-delete
	  denys::button-up-handler   'denys::button-nifty-ignore)))

;;--delete region. restore default state.
(defun denys::button-down-maybe-delete ()
  (if (not (= (% epoch::event-data 4) denys::nifty-select-button))
      (denys::mouse-ding)
    (delete-region (button-start denys::mouse-button)
		   (button-end   denys::mouse-button))
    (denys::warp-mouse-to-point)
    (denys::mouse-abort)))

;;--when nifty button is released, watch out for next click requesting
;;--copy of the region.
(defun denys::button-up-select ()
  (if (= (% epoch::event-data 4) denys::nifty-select-button)
      (setq denys::button-down-handler 'denys::button-down-maybe-copy
	    denys::button-up-handler   'denys::mouse-abort
	    denys::motion-handler      'denys::mouse-restart)
    (denys::mouse-ding)))

;;--scrolling while selecting
(defvar denys::mouse-hscroll nil)
(defvar denys::mouse-vscroll nil)

;;--begin selecting a region (with highlighting).
(defun denys::motion-maybe-select ()
  (denys::mouse-select)
  (setq denys::mouse-button-requested t
	denys::mouse-before           nil
	denys::mouse-button-buffer    (current-buffer)
	denys::mouse-button-start     (point))
  (denys::mouse-set-point)
  (setq denys::mouse-button-end       (point)
	denys::button-down-handler    'denys::mouse-restart
	denys::button-up-handler      'denys::button-up-select
	denys::motion-handler         'denys::motion-select
	denys::focus-handler          'denys::noop
	denys::mouse-edges            (window-edges))
  (denys::move-mouse-button))

;;--idem, but initiated with focus event
(defun denys::focus-maybe-select ()
  (epoch::query-mouse)
  (if (eq epoch::event-window (selected-window))
      (denys::motion-maybe-select)
    (setq denys::mouse-button-requested t
	  denys::mouse-before           nil
	  denys::mouse-button-buffer    (current-buffer)
	  denys::mouse-button-start     (point)
	  denys::mouse-button-end       (point)
	  denys::mouse-edges            (window-edges))
    (denys::motion-select)))

;;--track the pointer and extend the region being selected correspondingly.
;;--also, if the mouse leaves the window, we should begin scrolling so as
;;--to uncover the text hidden to the side where the mouse exited.
(defun denys::motion-select ()
  (if (eq epoch::event-window (selected-window))
      (denys::move-mouse-button)
    (setq denys::button-up-handler   'denys::mouse-abort
	  denys::button-down-handler 'denys::mouse-abort
	  denys::motion-handler      'denys::motion-select-scroll
	  denys::timer-handler       'denys::timer-select-scroll
	  denys::focus-handler       'denys::noop
	  denys::mouse-hscroll       nil
	  denys::mouse-vscroll       nil)
    (cond ((< epoch::event-y (nth 1 denys::mouse-edges))
	   (setq denys::mouse-vscroll 1))
	  ((>= epoch::event-y (1- (nth 3 denys::mouse-edges)))
	   (setq denys::mouse-vscroll -1))
	  ((< epoch::event-x (nth 0 denys::mouse-edges))
	   (setq denys::mouse-hscroll 1))
	  ((>= epoch::event-x (nth 2 denys::mouse-edges))
	   (setq denys::mouse-hscroll -1)))))

;;--how long to wait before we scroll another line
(defvar denys::mouse-scroll-delay 20
  "*how long to wait before we scroll another line. this is expressed
in milliseconds. it is difficult to ascribe precise meaning to this
number. just see what works best on your machine.")

;;--scroll for each timer event
(defun denys::timer-select-scroll ()
  (denys::mouse-select-scroll)
  (denys::mouse-timeout denys::mouse-scroll-delay))

(defun denys::motion-select-scroll ()
  (if (eq epoch::event-window (selected-window))
      (progn
	(denys::mouse-timeout 0)
	(setq denys::button-up-handler 'denys::button-up-select
	      denys::motion-handler    'denys::motion-select
	      denys::timer-handler     'denys::timer-normal)
	(denys::move-mouse-button))
    (denys::mouse-select-scroll)
    (denys::mouse-timeout denys::mouse-scroll-delay)))

(defun denys::mouse-select-scroll ()
  (cond (denys::mouse-hscroll
	 (scroll-right denys::mouse-hscroll)
	 (setq denys::tmp0 (window-hscroll))
	 (if (> denys::mouse-hscroll 0)
	     (move-to-column (1- denys::tmp0))
	   (move-to-column (+ (1- (window-width))
			      (if (> denys::tmp0 1) (1- denys::tmp0) denys::tmp0))))
	 (denys::move-mouse-button-to-point))
	(denys::mouse-vscroll
	 (scroll-down denys::mouse-vscroll)
	 (if (> denys::mouse-vscroll 0)
	     (goto-char (window-start))
	   (move-to-window-line (- (window-height) 2)))
	 (denys::move-mouse-button-to-point))))

;;--install nifty select on requested button
(defun denys::install-nifty-select (button)
  (aset denys::button-down-map button 'denys::mouse-begin-nifty-select)
  (aset denys::button-up-map   button 'denys::noop))

;;--------------------------------------------------------------------------------
;; MOTION HANDLER
;;--------------------------------------------------------------------------------

;;--handy function that highlights the line the mouse is on. skips
;;--leading white space.

(defun denys::track-mouse-line ()
  (if (window-coordinates epoch::event-x epoch::event-y
			  epoch::event-window nil
			  'denys::tmp0 'denys::tmp1)
      (progn
	(move-to-window-line denys::tmp1 epoch::event-window)
	(setq denys::mouse-button-requested t
	      denys::mouse-button-start (window-skip-chars-forward " \t" nil epoch::event-window)
	      denys::mouse-button-end (window-end-of-line 1 epoch::event-window)
	      denys::mouse-buffer (window-buffer epoch::event-window)
	      denys::tmp2 (window-hscroll epoch::event-window))
	(move-to-column (if (> denys::tmp2 1) (+ denys::tmp0 (1- denys::tmp2)) denys::tmp0)
			epoch::event-window))))

;;--some buffers may request that the mouse execute a particular
;;--procedure when the mouse is moving over it. typically, the task is
;;--to track the mouse and highlight the line it is on.

(defvar denys::motion-normal-alist
  '((dired-mode       . denys::track-mouse-line)
    (Buffer-menu-mode . denys::track-mouse-line)
    (index-mode       . denys::track-mouse-line)
    (rmail-summary    . denys::track-mouse-line))
  "*alist of elements of the form (major-mode . proc) where major-mode is a
symbol naming a major mode and proc is a procedure of no arguments to
be executed when the mouse is moving over a buffer with the corresponding
major mode. typically, this involves tracking the mouse and highlighting
the line it moves over.")
  
(defun denys::motion-normal ()
  (if (setq denys::tmp0 (assq (buffer-major-mode (window-buffer epoch::event-window))
			      denys::motion-normal-alist))
      (funcall (cdr denys::tmp0))))

;;--------------------------------------------------------------------------------
;; TIMER HANDLER
;;--------------------------------------------------------------------------------

;;--timeouts are requested by setting this variable. the post event handler
;;--honors the last posted request. once again we get request compression.
(defvar denys::mouse-timeout-request nil)

(defun denys::mouse-timeout-hook ()
  (if (and (numberp denys::mouse-timeout-request)
	   (>= denys::mouse-timeout-request 0))
      (progn
	(epoch::set-timer denys::mouse-timeout-request)
	(setq denys::mouse-timeout-request nil))))

(denys::hang-off-hook 'denys::post-event-handler-hook 'denys::mouse-timeout-hook)

(defun denys::mouse-timeout (timeout)
  (setq denys::mouse-timeout-request timeout))

;;--by default don't do anything
(defun denys::timer-normal () t)

;;--------------------------------------------------------------------------------
;; FOCUS HANDLER
;;--------------------------------------------------------------------------------

;;--on focus in, select new screen, maybe raise it.
(defun denys::focus-normal ()
  (if epoch::event-data
      (progn (select-screen epoch::event-screen)
	     (if auto-raise-screen (raise-screen epoch::event-screen)))))

;;--------------------------------------------------------------------------------
;; PROPERTY CHANGE HANDLER
;;--------------------------------------------------------------------------------

;;--default to normal epoch handler
(defun denys::property-change-normal ()
  (handle-property-change-events epoch::event-type
				 epoch::event-data
				 epoch::event-screen))

;;--------------------------------------------------------------------------------
;; CLIENT MESSAGE HANDLER
;;--------------------------------------------------------------------------------

(defvar denys::client-message-alist nil
  "*alist of elements of the form (x-resource . proc) where x-resource is
some x-resource (typically the result of a call epoch::intern-atom) and
proc is a procedure of 2 arguments DATA and SCREEN, where DATA is the
data correspnding to the client message, and SCREEN is the screen where
the event occurred.")

(defun denys::client-message-normal ()
  (if (setq denys::tmp0 (assq (car epoch::event-data) denys::client-message-alist))
      (funcall denys::tmp0 epoch::event-data epoch::event-screen)))

;;--------------------------------------------------------------------------------
;; MAP HANDLER
;;--------------------------------------------------------------------------------

(defun denys::map-normal () (rename-screen))

;;--------------------------------------------------------------------------------
;; MOVE HANDLER
;;--------------------------------------------------------------------------------

(defun denys::move-normal () t)

;;--------------------------------------------------------------------------------
;; RESIZE HANDLER
;;--------------------------------------------------------------------------------

(defun denys::resize-normal ()
  (epoch::redisplay-screen epoch::event-screen))

;;--------------------------------------------------------------------------------
;; AIM SCROLL
;;--------------------------------------------------------------------------------

;;--this implements continuous scrolling initiated by pressing a button and
;;--moving the mouse a few characters in the desired direction.

(defvar denys::mouse-aim-start-x nil)
(defvar denys::mouse-aim-start-y nil)
(defvar denys::mouse-aim-slippage  5)
(defvar denys::mouse-aim-hscroll nil)
(defvar denys::mouse-aim-vscroll nil)
(defvar denys::mouse-aim-window  nil)

;;--handy function. computes the absolute value of its argument.
(defun denys::abs (x) (if (< x 0) (- x) x))

;;--scroll window under the mouse in requested direction. note that said
;;--direction may vary dynamically.

(defun denys::mouse-aim-scroll ()
  (setq denys::tmp0 (if epoch::event-x (- epoch::event-x denys::mouse-aim-start-x) -1))
  (setq denys::mouse-aim-hscroll
	(cond ((< (denys::abs denys::tmp0) denys::mouse-aim-slippage) nil)
	      ((< denys::tmp0 0) -1)
	      (t 1)))
  (setq denys::tmp0 (if epoch::event-y (- epoch::event-y denys::mouse-aim-start-y) -1))
  (setq denys::mouse-aim-vscroll
	(cond ((< (denys::abs denys::tmp0) denys::mouse-aim-slippage) nil)
	      ((< denys::tmp0 0) -1)
	      (t 1)))
  (denys::scroll-window denys::mouse-aim-hscroll
			denys::mouse-aim-vscroll
			denys::mouse-aim-window)
  (denys::mouse-timeout denys::mouse-scroll-delay))

;;--scroll specified window both horizontally and vertically
;;--by requested amounts.
(defun denys::scroll-window (hscroll vscroll window)
  (if hscroll (scroll-right hscroll window))
  (if vscroll (scroll-down  vscroll window)))

;;--for each timer event scroll the window. the reason we need timer events
;;--is that typically the user moves the mouse a few characters away and
;;--then waits for the window to keep rolling. since no motion events are
;;--coming in, we need another way to periodically wake up and scroll. that
;;--is what timer events provide.

(defun denys::timer-aim-scroll ()
  (epoch::query-mouse)
  (denys::mouse-aim-scroll))

(defvar denys::aim-button nil)

(defun denys::aim-ignore ()
  (if (= (% epoch::event-data 4) denys::aim-button)
      (denys::mouse-abort)
    (denys::mouse-ding)))

;;--initiate continuous scrolling
(defun denys::mouse-aim-scroll-down ()
  (setq denys::aim-button (% epoch::event-data 4))
  (denys::mouse-abort)
  (setq denys::button-down-handler 'denys::mouse-ding
	denys::button-up-handler   'denys::aim-ignore
	denys::motion-handler      'denys::mouse-aim-scroll
	denys::timer-handler       'denys::timer-aim-scroll
	denys::focus-handler       'denys::noop
	denys::mouse-aim-start-x    epoch::event-x
	denys::mouse-aim-start-y    epoch::event-y
	denys::mouse-aim-hscroll    nil
	denys::mouse-aim-vscroll    nil
	denys::mouse-aim-window     epoch::event-window)
  (denys::mouse-timeout denys::mouse-scroll-delay))

(defun denys::install-aim-scroll (button)
  (aset denys::button-down-map button 'denys::mouse-aim-scroll-down)
  (aset denys::button-up-map   button 'denys::noop))
